'''
Created on Nov 27, 2012

@author: samagraw

Copyright (c) 2012-2013 by Cisco Systems, Inc.
All rights reserved.
'''

import logging
import json
import falcon

from apiservice import ResourceRoute
from common import AuthenticatedResource, make_response, make_error_response
from jsonencoder import JSONEncoder
from  logsinfo import LogsInfo
from ..utils.infraexceptions import C3Exception, MandatoryFileMissingError

jsonencoder = JSONEncoder()

log = logging.getLogger("runtime.api.resources")

@ResourceRoute("/platform/logs", endpoint="platformlogs")
class LogsListResourceHandler(AuthenticatedResource):

    def on_get(self, request, response):
        file_name = request.get_param("file")
        out = ''

        if file_name:
            try:
                out = LogsInfo.get_logs(file_name)
                response.set_header("Content-Type", "text/plain")
            except Exception as ex:
                if isinstance(ex, C3Exception):
                    response = make_error_response(response,
                                                    ex.message,
                                                    "File not found",
                                                    falcon.HTTP_404,
                                                    ex.errorcode)
                else:
                    response = make_error_response(response,
                                                    "File not found",
                                                    "File not found",
                                                    falcon.HTTP_404)
                return
        else:
            out = jsonencoder.encode(LogsInfo.get_logs_list(request.url))
            response.set_header("Content-Type", "application/json")

        response = make_response(response, out, falcon.HTTP_200)
        response.set_header("Cache-Control", "no-cache")


@ResourceRoute("/platform/taillog", endpoint="platformlogs")
class TaillogResourceHandler(AuthenticatedResource):

    def on_get(self, request, response):
        file_name = request.get_param("filename", required=True)
        lines = request.get_param("lines", required=False, default=10)
        try:
            out = LogsInfo.get_taillog(file_name, lines)
            out = jsonencoder.encode(out)
        except Exception as ex:
            if isinstance(ex, C3Exception):
                if isinstance(ex, MandatoryFileMissingError):
                    response = make_error_response(response,
                                                    "Given file %s is not found/invalid!"%file_name,
                                                    "Given file is not found/invalid!",
                                                    falcon.HTTP_404)
                    return
                response = make_error_response(response,
                                                ex.message,
                                                "Error while tailling the file %s. Cause %s"%(file_name, ex.message),
                                                falcon.HTTP_500,
                                                ex.errorcode)
                return
            elif isinstance(ex, IOError):
                response = make_error_response(response,
                                                "Given file %s is invalid!"%file_name,
                                                "Given file is invalid!",
                                                falcon.HTTP_415)
                return
            else:
                response = make_error_response(response,
                                                "Error while tailling the file %s. Cause %s"%(file_name, ex.message),
                                                "Error while tailling the file %s."%(file_name),
                                                falcon.HTTP_500)
            return


        response = make_response(response, out, falcon.HTTP_200)
        response.set_header("Cache-Control", "no-cache")


@ResourceRoute("/platform/loglevel", endpoint="platformloglevel")
class LogLevelResourceHandler(AuthenticatedResource):
   
    def on_get(self, request, response):
        defaultValue = request.get_param("default")
        if defaultValue=="true":
            out = jsonencoder.encode(LogsInfo.getDefaultLogLevel())
        else:
            out = jsonencoder.encode(LogsInfo.getRuntimeLogLevel())

        response = make_response(response,
                                 out,
                                 falcon.HTTP_200,
                                 {
                                     'Content-Type': "application/json",
                                     'Cache-Control': "no-cache"
                                 })


    def on_put(self, request, response):
        data = request.stream.read()
        if data == '':
            log.error("Request body for setting log level is empty")
            response = make_error_response(response,
                                            "Request body for setting log level is empty",
                                            "Request body not found",
                                            falcon.HTTP_404)
            return

        log.debug("Request data : %s" % str(data))
        data = json.loads(data)
        log.debug("Set log-level request recvd with data: %s", data)
        if LogsInfo.setLogLevel(data) == True:
            response = make_response(response, '',
                                     falcon.HTTP_200,
                                     {"Content-Type": "application/json"})
            return
        else:
            log.error("Invalid data entered to set log-levels: %s", data)
            response = make_error_response(response,
                                            "Invalid data entered to set log-levels: %s" % data,
                                            "Invalid logging management data",
                                            falcon.HTTP_404)
